@@ -2,12 +2,12 @@ from __future__ import annotations
 
 import functools
 import ipaddress
+import logging
 import os
 import string
 import sys
 import traceback
 import types
-import typing as t
 import warnings
 from collections import Counter, UserString
 from typing import (
@@ -21,7 +21,6 @@ from typing import (
     Mapping,
     MutableMapping,
     NamedTuple,
-    Tuple,
     TypeVar,
     Union,
     overload,
@@ -30,6 +29,8 @@ from urllib.parse import quote as urlquote
 from urllib.parse import urlsplit, urlunsplit
 
 import markdown
+import pathspec
+import pathspec.gitignore
 
 from mkdocs import plugins, theme, utils
 from mkdocs.config.base import (
@@ -59,9 +60,12 @@ class SubConfig(Generic[SomeConfig], BaseConfigOption[SomeConfig]):
     enable validation with `validate=True`.
     """
 
+    _config_file_path: str | None = None
+    config_class: type[SomeConfig]
+
     @overload
     def __init__(
-        self: SubConfig[SomeConfig], config_class: t.Type[SomeConfig], *, validate: bool = True
+        self: SubConfig[SomeConfig], config_class: type[SomeConfig], *, validate: bool = True
     ):
         """Create a sub-config in a type-safe way, using fields defined in a Config subclass."""
 
@@ -76,29 +80,37 @@ class SubConfig(Generic[SomeConfig], BaseConfigOption[SomeConfig]):
     def __init__(self, *config_options, validate=None):
         super().__init__()
         self.default = {}
-        if (
-            len(config_options) == 1
-            and isinstance(config_options[0], type)
-            and issubclass(config_options[0], Config)
-        ):
-            if validate is None:
-                validate = True
-            (self._make_config,) = config_options
-        else:
-            self._make_config = functools.partial(LegacyConfig, config_options)
-        self._do_validation = bool(validate)
+        self._do_validation = True if validate is None else validate
+        if type(self) is SubConfig:
+            if (
+                len(config_options) == 1
+                and isinstance(config_options[0], type)
+                and issubclass(config_options[0], Config)
+            ):
+                (self.config_class,) = config_options
+            else:
+                self.config_class = functools.partial(LegacyConfig, config_options)
+                self._do_validation = False if validate is None else validate
+
+    def __class_getitem__(cls, config_class: type[Config]):
+        """Eliminates the need to write `config_class = FooConfig` when subclassing SubConfig[FooConfig]"""
+        name = f'{cls.__name__}[{config_class.__name__}]'
+        return type(name, (cls,), dict(config_class=config_class))
+
+    def pre_validation(self, config: Config, key_name: str):
+        self._config_file_path = config.config_file_path
 
     def run_validation(self, value: object) -> SomeConfig:
-        config = self._make_config()
+        config = self.config_class(config_file_path=self._config_file_path)
         try:
-            config.load_dict(value)
+            config.load_dict(value)  # type: ignore
             failed, warnings = config.validate()
         except ConfigurationError as e:
             raise ValidationError(str(e))
 
         if self._do_validation:
             # Capture errors and warnings
-            self.warnings = [f"Sub-option '{key}': {msg}" for key, msg in warnings]
+            self.warnings.extend(f"Sub-option '{key}': {msg}" for key, msg in warnings)
             if failed:
                 # Get the first failing one
                 key, err = failed[0]
@@ -107,6 +119,28 @@ class SubConfig(Generic[SomeConfig], BaseConfigOption[SomeConfig]):
         return config
 
 
+class PropagatingSubConfig(SubConfig[SomeConfig], Generic[SomeConfig]):
+    """A SubConfig that must consist of SubConfigs with defined schemas.
+
+    Any value set on the top config gets moved to sub-configs with matching keys.
+    """
+
+    def run_validation(self, value: object):
+        if isinstance(value, dict):
+            to_discard = set()
+            for k1, v1 in self.config_class._schema:
+                if isinstance(v1, SubConfig):
+                    for k2, _ in v1.config_class._schema:
+                        if k2 in value:
+                            subdict = value.setdefault(k1, {})
+                            if isinstance(subdict, dict):
+                                to_discard.add(k2)
+                                subdict.setdefault(k2, value[k2])
+            for k in to_discard:
+                del value[k]
+        return super().run_validation(value)
+
+
 class OptionallyRequired(Generic[T], BaseConfigOption[T]):
     """
     Soft-deprecated, do not use.
@@ -155,7 +189,7 @@ class ListOfItems(Generic[T], BaseConfigOption[List[T]]):
     E.g. for `config_options.ListOfItems(config_options.Type(int))` a valid item is `[1, 2, 3]`.
     """
 
-    required: Union[bool, None] = None  # Only for subclasses to set.
+    required: bool | None = None  # Only for subclasses to set.
 
     def __init__(self, option_type: BaseConfigOption[T], default=None) -> None:
         super().__init__()
@@ -170,7 +204,7 @@ class ListOfItems(Generic[T], BaseConfigOption[List[T]]):
         self._config = config
         self._key_name = key_name
 
-    def run_validation(self, value: object) -> List[T]:
+    def run_validation(self, value: object) -> list[T]:
         if value is None:
             if self.required or self.default is None:
                 raise ValidationError("Required configuration not provided.")
@@ -191,6 +225,7 @@ class ListOfItems(Generic[T], BaseConfigOption[List[T]]):
         fake_keys = [f'{parent_key_name}[{i}]' for i in range(len(value))]
         fake_config.data = dict(zip(fake_keys, value))
 
+        self.option_type.warnings = self.warnings
         for key_name in fake_config:
             self.option_type.pre_validation(fake_config, key_name)
         for key_name in fake_config:
@@ -202,6 +237,63 @@ class ListOfItems(Generic[T], BaseConfigOption[List[T]]):
         return [fake_config[k] for k in fake_keys]
 
 
+class DictOfItems(Generic[T], BaseConfigOption[Dict[str, T]]):
+    """
+    Validates a dict of items. Keys are always strings.
+
+    E.g. for `config_options.DictOfItems(config_options.Type(int))` a valid item is `{"a": 1, "b": 2}`.
+    """
+
+    required: bool | None = None  # Only for subclasses to set.
+
+    def __init__(self, option_type: BaseConfigOption[T], default=None) -> None:
+        super().__init__()
+        self.default = default
+        self.option_type = option_type
+        self.option_type.warnings = self.warnings
+
+    def __repr__(self) -> str:
+        return f"{type(self).__name__}: {self.option_type}"
+
+    def pre_validation(self, config: Config, key_name: str):
+        self._config = config
+        self._key_name = key_name
+
+    def run_validation(self, value: object) -> Dict[str, T]:
+        if value is None:
+            if self.required or self.default is None:
+                raise ValidationError("Required configuration not provided.")
+            value = self.default
+        if not isinstance(value, dict):
+            raise ValidationError(f"Expected a dict of items, but a {type(value)} was given.")
+        if not value:  # Optimization for empty list
+            return value
+
+        fake_config = LegacyConfig(())
+        try:
+            fake_config.config_file_path = self._config.config_file_path
+        except AttributeError:
+            pass
+
+        # Emulate a config-like environment for pre_validation and post_validation.
+        fake_config.data = value
+
+        for key in fake_config:
+            self.option_type.pre_validation(fake_config, key)
+        for key in fake_config:
+            if not isinstance(key, str):
+                raise ValidationError(
+                    f"Expected type: {str} for keys, but received: {type(key)} (key={key})"
+                )
+        for key in fake_config:
+            # Specifically not running `validate` to avoid the OptionallyRequired effect.
+            fake_config[key] = self.option_type.run_validation(fake_config[key])
+        for key in fake_config:
+            self.option_type.post_validation(fake_config, key)
+
+        return value
+
+
 class ConfigItems(ListOfItems[LegacyConfig]):
     """
     Deprecated: Use `ListOfItems(SubConfig(...))` instead of `ConfigItems(...)`.
@@ -232,11 +324,11 @@ class Type(Generic[T], OptionallyRequired[T]):
     """
 
     @overload
-    def __init__(self, type_: t.Type[T], length: t.Optional[int] = None, **kwargs):
+    def __init__(self, type_: type[T], length: int | None = None, **kwargs):
         ...
 
     @overload
-    def __init__(self, type_: Tuple[t.Type[T], ...], length: t.Optional[int] = None, **kwargs):
+    def __init__(self, type_: tuple[type[T], ...], length: int | None = None, **kwargs):
         ...
 
     def __init__(self, type_, length=None, **kwargs) -> None:
@@ -265,7 +357,7 @@ class Choice(Generic[T], OptionallyRequired[T]):
     Validate the config option against a strict set of values.
     """
 
-    def __init__(self, choices: Collection[T], default: t.Optional[T] = None, **kwargs) -> None:
+    def __init__(self, choices: Collection[T], default: T | None = None, **kwargs) -> None:
         super().__init__(default=default, **kwargs)
         try:
             length = len(choices)
@@ -297,10 +389,10 @@ class Deprecated(BaseConfigOption):
 
     def __init__(
         self,
-        moved_to: t.Optional[str] = None,
-        message: t.Optional[str] = None,
+        moved_to: str | None = None,
+        message: str | None = None,
         removed: bool = False,
-        option_type: t.Optional[BaseConfigOption] = None,
+        option_type: BaseConfigOption | None = None,
     ) -> None:
         super().__init__()
         self.default = None
@@ -464,7 +556,7 @@ class Optional(Generic[T], BaseConfigOption[Union[T, None]]):
     def pre_validation(self, config: Config, key_name: str):
         return self.option.pre_validation(config, key_name)
 
-    def run_validation(self, value: object) -> Union[T, None]:
+    def run_validation(self, value: object) -> T | None:
         if value is None:
             return None
         return self.option.validate(value)
@@ -559,7 +651,7 @@ class EditURITemplate(BaseConfigOption[str]):
         def format(self, path, path_noext):
             return self.formatter.format(self.data, path=path, path_noext=path_noext)
 
-    def __init__(self, edit_uri_key: t.Optional[str] = None) -> None:
+    def __init__(self, edit_uri_key: str | None = None) -> None:
         super().__init__()
         self.edit_uri_key = edit_uri_key
 
@@ -610,7 +702,7 @@ class FilesystemObject(Type[str]):
     def __init__(self, exists: bool = False, **kwargs) -> None:
         super().__init__(type_=str, **kwargs)
         self.exists = exists
-        self.config_dir: t.Optional[str] = None
+        self.config_dir: str | None = None
 
     def pre_validation(self, config: Config, key_name: str):
         self.config_dir = (
@@ -639,7 +731,7 @@ class Dir(FilesystemObject):
 
 class DocsDir(Dir):
     def post_validation(self, config: Config, key_name: str):
-        if config.config_file_path is None:
+        if not config.config_file_path:
             return
 
         # Validate that the dir is not the parent dir of the config file.
@@ -757,7 +849,6 @@ class Theme(BaseConfigOption[theme.Theme]):
 
         # Ensure custom_dir is an absolute path
         if 'custom_dir' in theme_config and not os.path.isabs(theme_config['custom_dir']):
-            assert self.config_file_path is not None
             config_dir = os.path.dirname(self.config_file_path)
             theme_config['custom_dir'] = os.path.join(config_dir, theme_config['custom_dir'])
 
@@ -823,18 +914,51 @@ class Nav(OptionallyRequired):
             return f"a {type(value).__name__}: {value!r}"
 
 
-class Private(BaseConfigOption):
-    """
-    Private Config Option
+class Private(Generic[T], BaseConfigOption[T]):
+    """A config option that can only be populated programmatically. Raises an error if set by the user."""
 
-    A config option only for internal use. Raises an error if set by the user.
-    """
-
-    def run_validation(self, value: object):
+    def run_validation(self, value: object) -> None:
         if value is not None:
             raise ValidationError('For internal use only.')
 
 
+class ExtraScriptValue(Config):
+    """An extra script to be added to the page. The `extra_javascript` config is a list of these."""
+
+    path = Type(str)
+    """The value of the `src` tag of the script."""
+    type = Type(str, default='')
+    """The value of the `type` tag of the script."""
+    defer = Type(bool, default=False)
+    """Whether to add the `defer` tag to the script."""
+    async_ = Type(bool, default=False)
+    """Whether to add the `async` tag to the script."""
+
+    def __init__(self, path: str = '', config_file_path=None):
+        super().__init__(config_file_path=config_file_path)
+        self.path = path
+
+    def __str__(self):
+        return self.path
+
+    def __fspath__(self):
+        return self.path
+
+
+class ExtraScript(BaseConfigOption[Union[ExtraScriptValue, str]]):
+    def __init__(self):
+        super().__init__()
+        self.option_type = SubConfig[ExtraScriptValue]()
+
+    def run_validation(self, value: object) -> ExtraScriptValue | str:
+        self.option_type.warnings = self.warnings
+        if isinstance(value, str):
+            if value.endswith('.mjs'):
+                return self.option_type.run_validation({'path': value, 'type': 'module'})
+            return value
+        return self.option_type.run_validation(value)
+
+
 class MarkdownExtensions(OptionallyRequired[List[str]]):
     """
     Markdown Extensions Config Option
@@ -848,16 +972,16 @@ class MarkdownExtensions(OptionallyRequired[List[str]]):
 
     def __init__(
         self,
-        builtins: t.Optional[List[str]] = None,
+        builtins: list[str] | None = None,
         configkey: str = 'mdx_configs',
-        default: List[str] = [],
+        default: list[str] = [],
         **kwargs,
     ) -> None:
         super().__init__(default=default, **kwargs)
         self.builtins = builtins or []
         self.configkey = configkey
 
-    def validate_ext_cfg(self, ext, cfg):
+    def validate_ext_cfg(self, ext: object, cfg: object) -> None:
         if not isinstance(ext, str):
             raise ValidationError(f"'{ext}' is not a valid Markdown Extension name.")
         if not cfg:
@@ -866,8 +990,8 @@ class MarkdownExtensions(OptionallyRequired[List[str]]):
             raise ValidationError(f"Invalid config options for Markdown Extension '{ext}'.")
         self.configdata[ext] = cfg
 
-    def run_validation(self, value: object):
-        self.configdata: Dict[str, dict] = {}
+    def run_validation(self, value: object) -> list[str]:
+        self.configdata: dict[str, dict] = {}
         if not isinstance(value, (list, tuple, dict)):
             raise ValidationError('Invalid Markdown Extensions configuration')
         extensions = []
@@ -921,12 +1045,12 @@ class Plugins(OptionallyRequired[plugins.PluginCollection]):
     initializing the plugin class.
     """
 
-    def __init__(self, theme_key: t.Optional[str] = None, **kwargs) -> None:
+    def __init__(self, theme_key: str | None = None, **kwargs) -> None:
         super().__init__(**kwargs)
         self.installed_plugins = plugins.get_plugins()
         self.theme_key = theme_key
-        self._config: t.Optional[Config] = None
-        self.plugin_cache: Dict[str, plugins.BasePlugin] = {}
+        self._config: Config | None = None
+        self.plugin_cache: dict[str, plugins.BasePlugin] = {}
 
     def pre_validation(self, config, key_name):
         self._config = config
@@ -941,7 +1065,7 @@ class Plugins(OptionallyRequired[plugins.PluginCollection]):
         return self.plugins
 
     @classmethod
-    def _parse_configs(cls, value: Union[list, tuple, dict]) -> Iterator[Tuple[str, dict]]:
+    def _parse_configs(cls, value: list | tuple | dict) -> Iterator[tuple[str, dict]]:
         if isinstance(value, dict):
             for name, cfg in value.items():
                 if not isinstance(name, str):
@@ -960,7 +1084,7 @@ class Plugins(OptionallyRequired[plugins.PluginCollection]):
                     raise ValidationError(f"'{name}' is not a valid plugin name.")
                 yield name, cfg
 
-    def load_plugin_with_namespace(self, name: str, config) -> Tuple[str, plugins.BasePlugin]:
+    def load_plugin_with_namespace(self, name: str, config) -> tuple[str, plugins.BasePlugin]:
         if '/' in name:  # It's already specified with a namespace.
             # Special case: allow to explicitly skip namespaced loading:
             if name.startswith('/'):
@@ -1042,7 +1166,7 @@ class Hooks(BaseConfigOption[List[types.ModuleType]]):
     def run_validation(self, value: object) -> Mapping[str, Any]:
         paths = self._base_option.validate(value)
         self.warnings.extend(self._base_option.warnings)
-        value = t.cast(List[str], value)
+        assert isinstance(value, list)
 
         hooks = {}
         for name, path in zip(value, paths):
@@ -1067,3 +1191,31 @@ class Hooks(BaseConfigOption[List[types.ModuleType]]):
         plugins = config[self.plugins_key]
         for name, hook in config[key_name].items():
             plugins[name] = hook
+
+
+class PathSpec(BaseConfigOption[pathspec.gitignore.GitIgnoreSpec]):
+    """A path pattern based on gitignore-like syntax."""
+
+    def run_validation(self, value: object) -> pathspec.gitignore.GitIgnoreSpec:
+        if not isinstance(value, str):
+            raise ValidationError(f'Expected a multiline string, but a {type(value)} was given.')
+        try:
+            return pathspec.gitignore.GitIgnoreSpec.from_lines(lines=value.splitlines())
+        except ValueError as e:
+            raise ValidationError(str(e))
+
+
+class _LogLevel(OptionallyRequired[int]):
+    levels: Mapping[str, int] = {
+        "warn": logging.WARNING,
+        "info": logging.INFO,
+        "ignore": logging.DEBUG,
+    }
+
+    def run_validation(self, value: object) -> int:
+        if not isinstance(value, str):
+            raise ValidationError(f'Expected a string, but a {type(value)} was given.')
+        try:
+            return self.levels[value]
+        except KeyError:
+            raise ValidationError(f'Expected one of {list(self.levels)}, got {value!r}')
